#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <share.h>
#include <fcntl.h>
#include <io.h>
#include <dos.h>
#include "sf-nc.h"
#include "ndl.h"

FILE *handleNdl=NULL;

tNdlInfo NdlInfo[MAX_NODE_HDR];
tNodeHdrRec Header[ MAX_NODE_HDR ];
int ListCnt=0;

static char *E_lfn = "long file names (%s) not supported";
static char *E_open = "cannot open %.100s: %s";
static char *S_bp = "binary pattern";
static char *S_cp = "character pattern";
static char *S_el = "exclude like";
static char *S_io = "include only";
static char *S_cpf = "%-13s %s (%s)";
static char *S_bpf = "%-13s %d:%d/%d.%d (%s)";

#if defined(__386__)
    #define __Inline inline
#else
    #define __Inline
#endif

__Inline int isMaskedPattern(const char *pattern)
{
    if(strchr(pattern,'*'))return 1;
    if(strchr(pattern,'?'))return 1;
    if(strchr(pattern,'#'))return 1;
    return 0;
}

void GetQual( tAddressMask& Element, const char *key )
{
    Element.Masked = (unsigned char)isMaskedPattern(key);
    NdlInfo[ListCnt].Masked |= Element.Masked;
    if(Element.Masked)strcpy(Element.Mask.Pattern,key);
    else{
        setAddress( Element.Mask.Bin );
        Element.Mask.Bin.Zone = (short)atoi( key );
        key = strchr( key, ':' );
        if( key++ ){
            Element.Mask.Bin.Net = (short)atoi( key );
            key = strchr( key, '/' );
            if( key++ )Element.Mask.Bin.Node = (short)atoi( key );
        }
    }
}

void processList( char type, char *fnm, char *qual )
{
    char *key;

    if( ListCnt == MAX_NODE_HDR ){
        error( 2, "maximum %d nodelists allowed", MAX_NODE_HDR );
    }

    NdlInfo[ListCnt].ExcludeCnt = 0;
    NdlInfo[ListCnt].IncludeCnt = 0;
    NdlInfo[ListCnt].Masked = 0;
    NdlInfo[ListCnt].Crc = 0;
    NdlInfo[ListCnt].CfgLine = line;
    NdlInfo[ListCnt].DefZone=-1;
    NdlInfo[ListCnt].DefNet =-1;
    NdlInfo[ListCnt].DefRegion = 0;
    NdlInfo[ListCnt].Type=type;
    if( strlen(fnm) >= sizeof(NdlInfo[ListCnt].Name) )
        error( 2, E_lfn, fnm );
    strcpy( NdlInfo[ListCnt].Name, fnm );

    key = strtok( qual, token );
       
    while( key ){
        switch( key[0] ){
            case '+':
                if( NdlInfo[ListCnt].IncludeCnt == 8 ){
                    error( 2, "maximum include address mask number is 8" );
                }
                GetQual( NdlInfo[ListCnt].IncludeList[ NdlInfo[ListCnt].IncludeCnt ], key+1 );
                NdlInfo[ListCnt].IncludeCnt++;
                break;
            case '-':
                if( NdlInfo[ListCnt].ExcludeCnt == 8 ){
                    error( 2, "maximum exclude address mask number is 8" );
                }
                GetQual( NdlInfo[ListCnt].ExcludeList[ NdlInfo[ListCnt].ExcludeCnt ], key+1 );
                NdlInfo[ListCnt].ExcludeCnt++;
                break;
            case 'R':
            case 'r':
                if( type == LT_POINTLIST ){
                    error( 2, "default regions are only allowed in nodelists" );
                }
                key++;
                if( isdigit( key[0] ) ){
                    if( NdlInfo[ListCnt].DefRegion ){
                        error( 2, "doubly defined default region" );
                    }
                    NdlInfo[ListCnt].DefRegion = (short)atoi( key );
                    break;
                }
                goto UNREC;
            case '^':
                NdlInfo[ListCnt].Crc = 1;
                break;
            default:
                if( isdigit( key[0] ) ){
                    if( NdlInfo[ListCnt].DefZone != -1 ){
                        error( 2, "doubly defined default zone/net" );
                    }
                    NdlInfo[ListCnt].DefZone = (short)atoi( key );
                    key = strchr( key, ':' );
                    if( key++ )NdlInfo[ListCnt].DefNet = (short)atoi( key );
                    break;
                }
UNREC:          error( 2, "unrecognized qualifier %s", key );
        }
        key = strtok( NULL, token );
    }
    ListCnt++;
}

int GetNames(void)
{
    char filename[128];
    int i;
    unsigned char n;
    unsigned short date, time;
    FILE *handleIdx;
    tNodeIdxHdr Hdr;
    tNodeHdrRec HdrRec;
    int retval=1;

    strcpy( filename, nodepath );
    strcat( filename, "adphone.trx" );
    handleIdx = _fsopen( filename, "rb", SH_DENYWR );
    if( handleIdx )fclose( handleIdx );
    else retval=0;
    strcpy( filename, nodepath );
    strcat( filename, "adndl.idx" );
    handleIdx = _fsopen( filename, "rb", SH_DENYWR );
    errno = 0;
    if( handleIdx ){
        if( read( fileno(handleIdx), &Hdr, sizeof( Hdr ) ) < sizeof(Hdr))goto InvHdr;
        if( Hdr.TotalNdls != ListCnt ){
InvHdr:     fclose( handleIdx );
            retval = 0;
        }
    }else retval = 0;
    for( n=0; n<ListCnt; n++ ){
        line = NdlInfo[n].CfgLine;
        strcpy( filename, nodepath );
        strcat( filename, NdlInfo[ n ].Name );
        getLast( filename );
        for( i = strlen( filename ); i >= 0 && filename[i] != '\\'; i-- )continue;
        i++;
        Header[ n ].NameLen = (unsigned char)strlen( filename+i );
        if( Header[ n ].NameLen >= sizeof(NdlInfo[ n ].Name) ){
            if( handleIdx )fclose(handleIdx);
            error( 2, E_lfn, &filename[i] );
        }
        strncpy( Header[ n ].Name, filename+i, Header[ n ].NameLen );
        strcpy( NdlInfo[ n ].Name, filename+i );
        handleNdl = _fsopen( filename, "rb", SH_DENYWR );
        if( !handleNdl ){
            if( handleIdx )fclose(handleIdx);
            error( 3, E_open, filename, strerror(errno) );
        }
        if( _dos_getftime( fileno(handleNdl), &date, &time ) ){
            if( handleIdx )fclose(handleIdx);
            error( 3, "%s: %s", filename, strerror(errno) );
        }
        Header[ n ].Time = ((unsigned long)date) << 16 | time;
        fclose( handleNdl );
        handleNdl = NULL;
        if(handleIdx){
            if( read( fileno(handleIdx), &HdrRec, sizeof(tNodeHdrRec) ) < sizeof(tNodeHdrRec) )retval=0;
            if( memcmp( &HdrRec, &Header[ n ], sizeof(tNodeHdrRec) ) )retval = 0;
        }
    }
    if( handleIdx )fclose(handleIdx);
    return retval;
}

void CollectNdl(void)
{
    char filename[128];
    unsigned char n, i;

    for( n=0; n<ListCnt; n++ ){
        line = NdlInfo[ n ].CfgLine;
        strcpy( filename, nodepath );
        strcat( filename, NdlInfo[ n ].Name );
        plog( "incorporating %s %s", NdlInfo[ n ].Type==LT_NODELIST?"nodelist ":"pointlist", filename );
        if( NdlInfo[ n ].DefZone != -1 )plog( "assume default address %d:%d", (int)NdlInfo[ n ].DefZone, (int)NdlInfo[ n ].DefNet );
        if( NdlInfo[ n ].DefRegion )plog( "assume default region %d", (int)NdlInfo[ n ].DefRegion );
        for( i=0; i<NdlInfo[ n ].IncludeCnt; i++ ){
            if( NdlInfo[ n ].IncludeList[ i ].Masked )
                plog( S_cpf, i?"":S_io, NdlInfo[ n ].IncludeList[ i ].Mask.Pattern, S_cp );
            else
                plog( S_bpf, i?"":S_io,
                    (int)NdlInfo[ n ].IncludeList[ i ].Mask.Bin.Zone,
                    (int)NdlInfo[ n ].IncludeList[ i ].Mask.Bin.Net,
                    (int)NdlInfo[ n ].IncludeList[ i ].Mask.Bin.Node,
                    (int)NdlInfo[ n ].IncludeList[ i ].Mask.Bin.Point,
                    S_bp );
        }
        for( i=0; i<NdlInfo[ n ].ExcludeCnt; i++ ){
            if( NdlInfo[ n ].ExcludeList[ i ].Masked )
                plog( S_cpf, i?"":S_el, NdlInfo[ n ].ExcludeList[ i ].Mask.Pattern, S_cp );
            else
                plog( S_bpf, i?"":S_el,
                    (int)NdlInfo[ n ].ExcludeList[ i ].Mask.Bin.Zone,
                    (int)NdlInfo[ n ].ExcludeList[ i ].Mask.Bin.Net,
                    (int)NdlInfo[ n ].ExcludeList[ i ].Mask.Bin.Node,
                    (int)NdlInfo[ n ].ExcludeList[ i ].Mask.Bin.Point,
                    S_bp );
        }
        handleNdl = _fsopen( filename, "rb", SH_DENYWR );
        if( !handleNdl ){
            error( 3, E_open, filename, strerror(errno) );
        }
        collectRec( n );
        fclose( handleNdl );
        handleNdl = NULL;
    }
}
