#ifndef _LEX_CHECK_DEF

#define _LEX_CHECK_DEF

#ifndef LEX_MAX_STRING
    #define LEX_MAX_STRING 128
#endif

#pragma pack(1);
#pragma ZTC align 1

#ifdef __cplusplus
extern "C"{
#endif

typedef union{
            struct{
                unsigned int meaningBlanks          :1;
                unsigned int keywordCaseSensitive   :1;
                unsigned int wordCaseSensitive      :1;
                unsigned int symbolCaseSensitive    :1;
                unsigned int fullKeywords           :1;
                unsigned int noTruename             :1;
            }Flags;
            int FlagCluster;
}LEX_FLAGS;

typedef unsigned char LEX_SYNTAX;

typedef struct{
    unsigned char   sign;           // local state signature = 0
}LEX_STATE;

struct _LEX_TRAN;

struct _LEX_TRANSFER{
    char                    *rest;
    int                     length;
    char                    character;
    long                    integer;
    double                  real;
    unsigned char           lastype;
    LEX_SYNTAX *            syntax;
    LEX_STATE  *            source;
    LEX_STATE  *            destination;
    struct _LEX_TRAN *      translation;
    LEX_FLAGS               flags;
    char                    string[ LEX_MAX_STRING ];
};

typedef struct{
    unsigned char   type;
    LEX_STATE *     target;
    int cdecl (* routine)( struct _LEX_TRANSFER* );
}LEX_TRANREST;

struct _LEX_TRAN{
    LEX_TRANREST    tranrest;
    char            keyword;
};

typedef struct _LEX_TRAN LEX_TRAN;

typedef struct _LEX_TRANSFER LEX_TRANSFER;

#define LEX_EOS             1
#define LEX_BLANKS          2
#define LEX_NULL            3
#define LEX_ANY             4
#define LEX_ALPHA           5
#define LEX_WORD            6
#define LEX_DIGIT           7
#define LEX_ALNUM           8
#define LEX_SYMBOL          9
#define LEX_STRING          10
#define LEX_NUMBER          11
#define LEX_REAL            12
#define LEX_SIGN            13
#define LEX_LOGICAL         14
#define LEX_FNM             15
#define LEX_KEYWORD         16
#define LEX_SUBSYNTAX       33

#define LEX_FAIL            0xFFFF
#define LEX_SUCCESS         0
#define LEX_ERROR           1
#define LEX_E_UNKNOWNTYPE   10
#define LEX_E_EMPTYSTATE    11
#define LEX_E_CLINCH        12
#define LEX_E_MISMATCH      13
#define LEX_E_FAIL          14
#define LEX_E_ABBR          15

int _lexTranEos    ( LEX_TRANSFER* );
int _lexTranBlanks ( LEX_TRANSFER* );
int _lexTranNull   ( LEX_TRANSFER* );
int _lexTranAny    ( LEX_TRANSFER* );
int _lexTranAlpha  ( LEX_TRANSFER* );
int _lexTranWord   ( LEX_TRANSFER* );
int _lexTranDigit  ( LEX_TRANSFER* );
int _lexTranAlnum  ( LEX_TRANSFER* );
int _lexTranSymbol ( LEX_TRANSFER* );
int _lexTranString ( LEX_TRANSFER* );
int _lexTranNumber ( LEX_TRANSFER*, int );
int _lexTranReal   ( LEX_TRANSFER* );
int _lexTranSign   ( LEX_TRANSFER* );
int _lexTranLogical( LEX_TRANSFER* );
int _lexTranFnm    ( LEX_TRANSFER* );
int _lexTranKeyword( LEX_TRANSFER*, char * );

int lexical        ( LEX_SYNTAX *, LEX_FLAGS*, char* );
char *lexErrest    ( void );
char *lexError     ( void );
char * getcommand( void );

extern  int    lexErrno;
extern  char   *_lexErrstr;
extern  char   *_lexRest;

#define LEX_ROUTINE extern "C" int

#ifdef __cplusplus
    #define LEX_DECLARE_ROUTINE( routine ) extern "C" { extern int cdecl routine( LEX_TRANSFER * state ); }
    #define LEX_DECLARE_SYNTAX( syntax ) extern "C" { extern LEX_SYNTAX syntax; }
#else
    #define LEX_DECLARE_ROUTINE( routine ) extern int cdecl routine( LEX_TRANSFER * state );
    #define LEX_DECLARE_SYNTAX( syntax ) extern LEX_SYNTAX syntax;
#endif

#ifdef __cplusplus
}
#endif

#pragma ZTC align
#pragma pack();

#endif

