#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <dos.h>
#include <stdlib.h>
#include <malloc.h>
#include "TScreen.h"
#include "colors.h"
#include "mouse.h"
#include "key.h"

TScreen    Screen;

static unsigned char        TScreen::cursor = _CR_UNK;
static int                  TScreen::maxRow=0;
static int                  TScreen::maxCol=0;
static unsigned char far*   TScreen::screenAddr;
static unsigned char        TScreen::ega;
static unsigned char        TScreen::videoMode;
static unsigned char        TScreen::isColor;

TScreen::TScreen()
{
  REGS regs;
  x = y = 1;

  trackCursor = 0;
  scrstack = NULL;
  regs.h.ah = 0x0F;
  int86( 0x10, &regs, &regs );
  videoMode = regs.h.al;
  isColor = (videoMode!=7)&&(videoMode!=2);
  screenAddr = (unsigned char far*)(isColor?0xB8000000l:0xB0000000l);
  ptr = screenAddr;

  ega = * (unsigned char far*)MK_FP( 0x0040, 0x0087 );
  ega = (!( ega & 8 )) && ega;

  maxCol = * (unsigned char far*)MK_FP( 0, 0x044A );
  if( videoMode < 2 ){
    maxRow = 25;
  }else{
    if( !ega ){
        maxRow = 25;
    }else{
        maxRow = ( *(unsigned char far*)MK_FP( 0, 0x0484 ) ) + 1;
    }
  }

  background = WHITE;
  foreground = BLACK;
}

void TScreen::setxy( int newx, int newy )
{
  x = newx;
  y = newy;
//  ptr = (unsigned char far*)((unsigned char far*)screenAddr + (y-1)*maxCol*2 + (((unsigned long)x-1l)<<1));
  ptr = (unsigned char far*)((unsigned short far*)screenAddr + ((y-1)*maxCol + x-1));
  if(trackCursor)screen_setxy( newx, newy );
}

void TScreen::outc( char symbol )
{
  disableMouse();

  *ptr = symbol;
  ptr++;
  x++;
  *ptr = (unsigned char)((foreground&0x0F) + ((background&0x0F)<<4));
  ptr++;
  if(trackCursor)screen_setxy( x, y );

  enableMouse();
}

void TScreen::outs( char *string )
{
  register int i;

  mouseLongUpdate();
  for( i = 0; string[ i ]; i++ )TScreen::outc( string[ i ] );
  mouseEndUpdate();
}

void TScreen::clrArea( int left, int top, int right, int bottom, char filler  )
{
    int xc, yc, i, attr;

    attr = ((background&0x0F)<<4) | (foreground&0x0F);

    disableMouse();

    for( xc=top-1; xc<bottom; xc++ ){
        for( yc=left-1; yc<right; yc++){
            i = (xc*maxCol + yc)<<1;
            screenAddr[ i++ ] = filler;
            screenAddr[ i ] = (unsigned char)attr;
        }
    }

    enableMouse();
}

void TScreen::clrScreen( char filler )
{
    clrArea( 1, 1, maxCol, maxRow, filler );
}


void TScreen::outf( char *format, ... )
{
    va_list argptr;
    char str[256];

    va_start( argptr, format );
    vsprintf( str, format, argptr );
    va_end( argptr );
    outs( str );
}

void TScreen::scroll( int dir, int left, int top, int right, int bottom, int bkg, int clr )
{
    int i, bkg1, clr1;

    bkg1 = background;
    clr1 = foreground;
    background = bkg;
    foreground = clr;
    if(dir>0){
        for( ; top < bottom; top++ ){
            for( i=left-1; i<right; i++ ){
                ((short far*)screenAddr)[ (top-1)*maxCol + i ] = ((short far*)screenAddr)[ top*maxCol + i ];
            }
        }
        clrArea( left, bottom, right, bottom, ' ' );
    }else{
        for( ; bottom > top; bottom-- ){
            for( i=left-1; i<right; i++ ){
                ((short far*)screenAddr)[ (bottom-1)*maxCol + i ] = ((short far*)screenAddr)[ (bottom-2)*maxCol + i ];
            }
        }
        clrArea( left, top, right, top, ' ' );
    }
    background = bkg1;
    foreground = clr1;
}

void TScreen::screen_setxy(int x, int y)
{
#if defined(__TURBOC__)
    _AH = 0x02;
    _BH = 0;
    _DH = y - 1;
    _DL = x - 1;
    geninterrupt(0x10);
#else
    union REGS regs;
    regs.h.ah = 0x02;
    regs.h.bh = 0;
    regs.h.dh = (unsigned char)(y - 1);
    regs.h.dl = (unsigned char)(x - 1);
#if defined(__WATCOMC__) && defined(__386__)
    int386(0x10, &regs, &regs);
#else
    int86(0x10, &regs, &regs);
#endif
#endif
}

void TScreen::set_track_cursor(int b)
{
    trackCursor = (b!=0);
    if(b)screen_setxy(x,y);
}

static void TScreen::setcursorsize(char start, char end)
{
#if defined(__TURBOC__)
    _AH = 0x01;
    _CH = start;
    _CL = end;
    geninterrupt(0x10);
#else
    union REGS regs;
    regs.h.ah = 0x01;
    regs.h.ch = start;
    regs.h.cl = end;
#if defined(__WATCOMC__) && defined(__386__)
    int386(0x10, &regs, &regs);
#else
    int86(0x10, &regs, &regs);
#endif
#endif
}

static void TScreen::cursorblock(void)
{
    if( cursor != _CR_BLOCK ){
        if (isColor)setcursorsize(0, 7);
        else setcursorsize(0, 13);
        cursor = _CR_BLOCK;
    }
}

static void TScreen::cursornorm(void)
{
    if( cursor != _CR_NORM ){
        if (isColor)setcursorsize(7, 8);
        else setcursorsize(11, 13);
        cursor = _CR_NORM;
    }
}

static void TScreen::cursoroff(void)
{
    if( cursor != _CR_OFF ){
        setcursorsize(32, 32);
        cursor = _CR_OFF;
    }
}

static void TScreen::invertColor( int &back, int &text )
{
    int c = text;

    text = (back & 7) /* | (text & 8) */;
    back = (c    & 7) | (back & 8);
}

int TScreen::cgetDlg( char *str )
{
  int key, count = strlen( str+2 );
  char *oldstr;
  int isFirst = 1;
  int curpos = count;
  int i;
  int insmode = 1;
  char oldtrack = trackCursor;

  outs( str+2 );

  oldstr = (char*)alloca( count + 3 );
  strcpy( oldstr+2, str+2 );
  oldstr[ 0 ] = str[ 0 ];
  oldstr[ 1 ] = str[ 1 ];

  set_track_cursor(1);
  cursornorm();
  while( 1 ){
    switch( key = mgetch() ){
	  case Left:
	    if( curpos ){
	      curpos--;
          setxy( x-1, y );
	    }
	    break;
	  case Right:
	    if( curpos < count ){
	      curpos++;
          setxy( x+1, y );
	    }
	    break;
	  case Del:
DeleteLabel:
	    if( curpos < count ){
	      str[ count + 2 ] = NULL;
	      strcpy( str+curpos+2, str+curpos+3 );
	      count--;
          i = x;
          outs( str+curpos+2 );
          outc( ' ' );
          setxy( i, y );
	    }
	    break;
	  case Home:
        setxy( x - curpos, y );
	    curpos = 0;
	    break;
	  case End:
        setxy( x - curpos + count, y );
	    curpos = count;
	    break;
	  case Ins:
	    insmode^=1;
        if( insmode )cursornorm();
        else cursorblock();
	    break;
      case BackSpace:
        if( curpos != 0 ){
          if( curpos == count ){
            count--;
            curpos--;
            setxy( x-1, y );
            outc( ' ' );
            setxy( x-1, y );
          }else{
            curpos--;
            setxy( x-1, y );
            goto DeleteLabel;
          }
        }
        break;
      case Esc:
      case RMB:
        mouseWaitRelease();
        strcpy( str+2, oldstr+2 );
        str[ 1 ] = (char)strlen( str+2 );
        cursornorm();
        set_track_cursor(oldtrack);
        return 0;
      case Enter:
      case LMB:
        mouseWaitRelease();
EnterLabel:
        str[ 1 ] = (char)count;
        str[ count + 2 ] = NULL;
        cursornorm();
        set_track_cursor(oldtrack);
        return 1;
      default:
        if( key > 255 )break;
        if( curpos != count ){
          if( insmode ){
            for( i = count; i > curpos; i-- )str[ i+2 ] = str[ i+1 ];
            count++;
          }
          str[ count+2 ] = NULL;
          str[ ++curpos+1 ] = (char)key;
          i = x;
          outs( str+curpos+1 );
          setxy( i+1, y );
        }else{
          if( isFirst ){
            setxy( x - curpos, y );
            i = x;
            for( isFirst = 0; isFirst < count; isFirst++ )outc( ' ' );
            count = 0;
            curpos = 0;
            setxy( i, y );
          }
          if( count < str[ 0 ] ){
            str[ ++count+1 ] = (char)key;
            curpos++;
            outc( (char)key );
          }
        }
        if( curpos == str[ 0 ] )goto EnterLabel;
    }
    isFirst = 0;
  }
}

static void TScreen::selectLine( int left, int top, int right, int background, int text )
{
  int i,j;

  top--;
  i=((top*maxCol+left-1)<<1)+1;
  disableMouse();
  for( j=0; j< ((right-left)<<1); j+=2 ){
    screenAddr[ i+j ] = (unsigned char)((text&0x0F) + ((background&0x0F)<<4));
  }
  enableMouse();
}

static void far* TScreen::savescr( void far* buff )
{
    REGS regs;

    if(!buff)buff = _fmalloc( (maxCol*maxRow<<1)+6 );
    _heapgrow();
    disableMouse();

    _fmemcpy( buff, screenAddr, maxCol*maxRow<<1 );
    regs.h.ah = 0x03;
    regs.h.bh = 0;
    int86( 0x10, &regs, &regs );
    ((short far*)buff)[maxCol*maxRow+1] = regs.x.cx;
    ((short far*)buff)[maxCol*maxRow]   = regs.x.dx;
    enableMouse();
    return buff;
}

static void TScreen::restscr( void far* buff, int freemem )
{
    REGS regs;

    disableMouse();

    _fmemcpy( screenAddr, buff, maxCol*maxRow<<1 );
    regs.h.bh   = 0;
    regs.x.cx = ((short far*)buff)[maxCol*maxRow+1];
    regs.h.ah = 0x01;
    int86( 0x10, &regs, &regs );
    regs.x.dx = ((short far*)buff)[maxCol*maxRow];
    regs.h.bh   = 0;
    regs.h.ah = 0x02;
    int86( 0x10, &regs, &regs );
    enableMouse();
    if(freemem)_ffree( buff );
}

void TScreen::push(void)
{
    ScreenList *ptr;

    ptr = scrstack;
    scrstack = new ScreenList;
    scrstack->prev = ptr;
    scrstack->mem = savescr();
}

void TScreen::pop(void)
{
    ScreenList *ptr;

    if(!scrstack)return;
    ptr = scrstack->prev;
    restscr(scrstack->mem);
    delete scrstack;
    scrstack = ptr;
}

